@@ -1,6 +1,6 @@
 /*
  * Copyright (c) [2012-2015] Novell, Inc.
- * Copyright (c) [2016-2022] SUSE LLC
+ * Copyright (c) [2016-2023] SUSE LLC
  *
  * All Rights Reserved.
  *
@@ -91,7 +91,8 @@ Client::find_comparison(Snapper* snapper, Snapshots::const_iterator snapshot1,
 	    return it;
     }
 
-    throw NoComparison();
+    SN_THROW(NoComparison());
+    __builtin_unreachable();
 }
 
 
@@ -313,6 +314,18 @@ Client::introspect(DBus::Connection& conn, DBus::Message& msg)
 	"      <arg name='numbers' type='au' direction='in'/>\n"
 	"    </method>\n"
 
+	"    <method name='IsSnapshotReadOnly'>\n"
+	"      <arg name='config-name' type='s' direction='in'/>\n"
+	"      <arg name='number' type='u' direction='in'/>\n"
+	"      <arg name='read-only' type='b' direction='out'/>\n"
+	"    </method>\n"
+
+	"    <method name='SetSnapshotReadOnly'>\n"
+	"      <arg name='config-name' type='s' direction='in'/>\n"
+	"      <arg name='number' type='u' direction='in'/>\n"
+	"      <arg name='read-only' type='b' direction='in'/>\n"
+	"    </method>\n"
+
 	"    <method name='GetDefaultSnapshot'>\n"
 	"      <arg name='config-name' type='s' direction='in'/>\n"
 	"      <arg name='valid' type='b' direction='out'/>\n"
@@ -388,6 +401,8 @@ Client::introspect(DBus::Connection& conn, DBus::Message& msg)
 	"  </interface>\n"
 	"</node>\n";
 
+    y2deb("Introspect");
+
     DBus::MessageMethodReturn reply(msg);
 
     DBus::Marshaller marshaller(reply);
@@ -410,7 +425,7 @@ Client::check_permission(DBus::Connection& conn, DBus::Message& msg) const
     if (uid == 0)
 	return;
 
-    throw Permissions();
+    SN_THROW(Permissions());
 }
 
 
@@ -444,7 +459,7 @@ Client::check_permission(DBus::Connection& conn, DBus::Message& msg,
 		return;
     }
 
-    throw Permissions();
+    SN_THROW(Permissions());
 }
 
 
@@ -463,7 +478,7 @@ Client::check_lock(DBus::Connection& conn, DBus::Message& msg, const string& con
 	    continue;
 
 	if (it->has_lock(config_name))
-	    throw Lock();
+	    SN_THROW(Lock());
     }
 }
 
@@ -484,7 +499,7 @@ void
 Client::check_config_in_use(const MetaSnapper& meta_snapper) const
 {
     if (meta_snapper.use_count() != 0)
-	throw ConfigInUse();
+	SN_THROW(ConfigInUse());
 }
 
 
@@ -497,7 +512,7 @@ Client::check_snapshot_in_use(const MetaSnapper& meta_snapper, unsigned int numb
 	    it1->mounts.find(make_pair(meta_snapper.configName(), number));
 
 	if (it2 != it1->mounts.end())
-	    throw SnapshotInUse();
+	    SN_THROW(SnapshotInUse());
     }
 }
 
@@ -566,7 +581,7 @@ Client::signal_snapshot_modified(DBus::Connection& conn, const string& config_na
 
 void
 Client::signal_snapshots_deleted(DBus::Connection& conn, const string& config_name,
-				 const list<dbus_uint32_t>& nums)
+				 const vector<dbus_uint32_t>& nums)
 {
     DBus::MessageSignal msg(PATH, INTERFACE, "SnapshotsDeleted");
 
@@ -589,7 +604,7 @@ Client::list_configs(DBus::Connection& conn, DBus::Message& msg)
     DBus::Marshaller marshaller(reply);
     marshaller.open_array(DBus::TypeInfo<ConfigInfo>::signature);
     for (MetaSnappers::const_iterator it = meta_snappers.begin(); it != meta_snappers.end(); ++it)
-    marshaller << it->getConfigInfo();
+	marshaller << it->getConfigInfo();
     marshaller.close_array();
 
     conn.send(reply);
@@ -839,7 +854,7 @@ Client::get_snapshot(DBus::Connection& conn, DBus::Message& msg)
 
     Snapshots::iterator snap = snapshots.find(num);
     if (snap == snapshots.end())
-	throw IllegalSnapshotException();
+	SN_THROW(IllegalSnapshotException());
 
     DBus::MessageMethodReturn reply(msg);
 
@@ -873,7 +888,7 @@ Client::set_snapshot(DBus::Connection& conn, DBus::Message& msg)
 
     Snapshots::iterator snap = snapshots.find(num);
     if (snap == snapshots.end())
-	throw IllegalSnapshotException();
+	SN_THROW(IllegalSnapshotException());
 
     snapper->modifySnapshot(snap, smd);
 
@@ -1073,12 +1088,12 @@ void
 Client::delete_snapshots(DBus::Connection& conn, DBus::Message& msg)
 {
     string config_name;
-    list<dbus_uint32_t> nums;
+    vector<dbus_uint32_t> nums;
 
     DBus::Unmarshaller unmarshaller(msg);
     unmarshaller >> config_name >> nums;
 
-    y2deb("DeleteSnapshots config_name:" << config_name << " nums:" << nums);
+    y2mil("DeleteSnapshots config_name:" << config_name << " nums:" << nums);
 
     boost::unique_lock<boost::shared_mutex> lock(big_mutex);
 
@@ -1091,7 +1106,7 @@ Client::delete_snapshots(DBus::Connection& conn, DBus::Message& msg)
     Snapper* snapper = it1->getSnapper();
     Snapshots& snapshots = snapper->getSnapshots();
 
-    for (list<unsigned int>::const_iterator it2 = nums.begin(); it2 != nums.end(); ++it2)
+    for (vector<unsigned int>::const_iterator it2 = nums.begin(); it2 != nums.end(); ++it2)
     {
 	check_snapshot_in_use(*it1, *it2);
 
@@ -1108,6 +1123,76 @@ Client::delete_snapshots(DBus::Connection& conn, DBus::Message& msg)
 }
 
 
+void
+Client::is_snapshot_read_only(DBus::Connection& conn, DBus::Message& msg)
+{
+    string config_name;
+    dbus_uint32_t num;
+
+    DBus::Unmarshaller unmarshaller(msg);
+    unmarshaller >> config_name >> num;
+
+    y2deb("IsSnapshotReadOnly config_name:" << config_name << " num:" << num);
+
+    boost::unique_lock<boost::shared_mutex> lock(big_mutex);
+
+    MetaSnappers::iterator it = meta_snappers.find(config_name);
+
+    check_permission(conn, msg, *it);
+
+    Snapper* snapper = it->getSnapper();
+
+    Snapshots& snapshots = snapper->getSnapshots();
+
+    Snapshots::iterator snap = snapshots.find(num);
+    if (snap == snapshots.end())
+	SN_THROW(IllegalSnapshotException());
+
+    bool read_only = snap->isReadOnly();
+
+    DBus::MessageMethodReturn reply(msg);
+
+    DBus::Marshaller marshaller(reply);
+    marshaller << read_only;
+
+    conn.send(reply);
+}
+
+
+void
+Client::set_snapshot_read_only(DBus::Connection& conn, DBus::Message& msg)
+{
+    string config_name;
+    dbus_uint32_t num;
+    bool read_only;
+
+    DBus::Unmarshaller unmarshaller(msg);
+    unmarshaller >> config_name >> num >> read_only;
+
+    y2deb("SetSnapshotReadOnly config_name:" << config_name << " num:" << num << " read_only:" << read_only);
+
+    boost::unique_lock<boost::shared_mutex> lock(big_mutex);
+
+    MetaSnappers::iterator it = meta_snappers.find(config_name);
+
+    check_permission(conn, msg, *it);
+
+    Snapper* snapper = it->getSnapper();
+
+    Snapshots& snapshots = snapper->getSnapshots();
+
+    Snapshots::iterator snap = snapshots.find(num);
+    if (snap == snapshots.end())
+	SN_THROW(IllegalSnapshotException());
+
+    snap->setReadOnly(read_only);
+
+    DBus::MessageMethodReturn reply(msg);
+
+    conn.send(reply);
+}
+
+
 void
 Client::get_default_snapshot(DBus::Connection& conn, DBus::Message& msg)
 {
@@ -1134,9 +1219,9 @@ Client::get_default_snapshot(DBus::Connection& conn, DBus::Message& msg)
     DBus::Marshaller marshaller(reply);
 
     if (tmp != snapshots.end())
-    marshaller << true << tmp->getNum();
+	marshaller << true << tmp->getNum();
     else
-    marshaller << false << (unsigned int)(0);
+	marshaller << false << (unsigned int)(0);
 
     conn.send(reply);
 }
@@ -1168,9 +1253,9 @@ Client::get_active_snapshot(DBus::Connection& conn, DBus::Message& msg)
     DBus::Marshaller marshaller(reply);
 
     if (tmp != snapshots.end())
-    marshaller << true << tmp->getNum();
+	marshaller << true << tmp->getNum();
     else
-    marshaller << false << (unsigned int)(0);
+	marshaller << false << (unsigned int)(0);
 
     conn.send(reply);
 }
@@ -1225,7 +1310,7 @@ Client::get_used_space(DBus::Connection& conn, DBus::Message& msg)
 
     Snapshots::iterator snap = snapshots.find(num);
     if (snap == snapshots.end())
-	throw IllegalSnapshotException();
+	SN_THROW(IllegalSnapshotException());
 
     uint64_t used_space = snap->getUsedSpace();
 
@@ -1262,7 +1347,7 @@ Client::mount_snapshot(DBus::Connection& conn, DBus::Message& msg)
 
     Snapshots::iterator snap = snapshots.find(num);
     if (snap == snapshots.end())
-	throw IllegalSnapshotException();
+	SN_THROW(IllegalSnapshotException());
 
     snap->mountFilesystemSnapshot(user_request);
 
@@ -1304,7 +1389,7 @@ Client::umount_snapshot(DBus::Connection& conn, DBus::Message& msg)
 
     Snapshots::iterator snap = snapshots.find(num);
     if (snap == snapshots.end())
-	throw IllegalSnapshotException();
+	SN_THROW(IllegalSnapshotException());
 
     snap->umountFilesystemSnapshot(user_request);
 
@@ -1338,7 +1423,7 @@ Client::get_mount_point(DBus::Connection& conn, DBus::Message& msg)
     Snapshots& snapshots = snapper->getSnapshots();
     Snapshots::iterator snap = snapshots.find(num);
     if (snap == snapshots.end())
-	throw IllegalSnapshotException();
+	SN_THROW(IllegalSnapshotException());
 
     string mount_point = snap->snapshotDir();
 
@@ -1624,7 +1709,7 @@ Client::sync(DBus::Connection& conn, DBus::Message& msg)
 
 
 void
-Client::debug(DBus::Connection& conn, DBus::Message& msg) const
+Client::debug(DBus::Connection& conn, DBus::Message& msg)
 {
     y2deb("Debug");
 
@@ -1642,7 +1727,7 @@ Client::debug(DBus::Connection& conn, DBus::Message& msg) const
     {
 	std::ostringstream s;
 	s << "    pid:" << getpid();
-    marshaller << s.str();
+	marshaller << s.str();
     }
 
     marshaller << "clients:";
@@ -1658,7 +1743,7 @@ Client::debug(DBus::Connection& conn, DBus::Message& msg) const
 	    s << ", locks " << it->locks.size();
 	if (!it->comparisons.empty())
 	    s << ", comparisons " << it->comparisons.size();
-    marshaller << s.str();
+	marshaller << s.str();
     }
 
     marshaller << "backgrounds:";
@@ -1666,7 +1751,7 @@ Client::debug(DBus::Connection& conn, DBus::Message& msg) const
     {
 	std::ostringstream s;
 	s << "    name:'" << it->meta_snapper->configName() << "'";
-    marshaller << s.str();
+	marshaller << s.str();
     }
 
     marshaller << "meta-snappers:";
@@ -1682,7 +1767,7 @@ Client::debug(DBus::Connection& conn, DBus::Message& msg) const
 	    else
 		s << ", use count " << it->use_count();
 	}
-    marshaller << s.str();
+	marshaller << s.str();
     }
 
     marshaller << "compile options:";
@@ -1698,81 +1783,61 @@ Client::debug(DBus::Connection& conn, DBus::Message& msg) const
 void
 Client::dispatch(DBus::Connection& conn, DBus::Message& msg)
 {
+    using method_fnc = void (Client ::*)(DBus::Connection& conn, DBus::Message& msg);
+
+    static const vector<pair<const char*, method_fnc>> method_registry = {
+	{ "ListConfigs", &Client::list_configs },
+	{ "CreateConfig", &Client::create_config },
+	{ "GetConfig", &Client::get_config },
+	{ "SetConfig", &Client::set_config },
+	{ "DeleteConfig", &Client::delete_config },
+	{ "LockConfig", &Client::lock_config },
+	{ "UnlockConfig", &Client::unlock_config },
+	{ "ListSnapshots", &Client::list_snapshots },
+	{ "ListSnapshotsAtTime", &Client::list_snapshots_at_time },
+	{ "GetSnapshot", &Client::get_snapshot },
+	{ "SetSnapshot", &Client::set_snapshot },
+	{ "CreateSingleSnapshot", &Client::create_single_snapshot },
+	{ "CreateSingleSnapshotV2", &Client::create_single_snapshot_v2 },
+	{ "CreateSingleSnapshotOfDefault", &Client::create_single_snapshot_of_default },
+	{ "CreatePreSnapshot", &Client::create_pre_snapshot },
+	{ "CreatePostSnapshot", &Client::create_post_snapshot },
+	{ "DeleteSnapshots", &Client::delete_snapshots },
+	{ "IsSnapshotReadOnly", &Client::is_snapshot_read_only },
+	{ "SetSnapshotReadOnly", &Client::set_snapshot_read_only },
+	{ "GetDefaultSnapshot", &Client::get_default_snapshot },
+	{ "GetActiveSnapshot", &Client::get_active_snapshot },
+	{ "CalculateUsedSpace", &Client::calculate_used_space },
+	{ "GetUsedSpace", &Client::get_used_space },
+	{ "MountSnapshot", &Client::mount_snapshot },
+	{ "UmountSnapshot", &Client::umount_snapshot },
+	{ "GetMountPoint", &Client::get_mount_point },
+	{ "CreateComparison", &Client::create_comparison },
+	{ "DeleteComparison", &Client::delete_comparison },
+	{ "GetFiles", &Client::get_files },
+	{ "GetFilesByPipe", &Client::get_files_by_pipe },
+	{ "SetupQuota", &Client::setup_quota },
+	{ "PrepareQuota", &Client::prepare_quota },
+	{ "QueryQuota", &Client::query_quota },
+	{ "QueryFreeSpace", &Client::query_free_space },
+	{ "Sync", &Client::sync },
+	{ "Debug", &Client::debug }
+    };
+
     try
     {
-	if (msg.is_method_call(INTERFACE, "ListConfigs"))
-	    list_configs(conn, msg);
-	else if (msg.is_method_call(INTERFACE, "CreateConfig"))
-	    create_config(conn, msg);
-	else if (msg.is_method_call(INTERFACE, "GetConfig"))
-	    get_config(conn, msg);
-	else if (msg.is_method_call(INTERFACE, "SetConfig"))
-	    set_config(conn, msg);
-	else if (msg.is_method_call(INTERFACE, "DeleteConfig"))
-	    delete_config(conn, msg);
-	else if (msg.is_method_call(INTERFACE, "LockConfig"))
-	    lock_config(conn, msg);
-	else if (msg.is_method_call(INTERFACE, "UnlockConfig"))
-	    unlock_config(conn, msg);
-	else if (msg.is_method_call(INTERFACE, "ListSnapshots"))
-	    list_snapshots(conn, msg);
-	else if (msg.is_method_call(INTERFACE, "ListSnapshotsAtTime"))
-	    list_snapshots_at_time(conn, msg);
-	else if (msg.is_method_call(INTERFACE, "GetSnapshot"))
-	    get_snapshot(conn, msg);
-	else if (msg.is_method_call(INTERFACE, "SetSnapshot"))
-	    set_snapshot(conn, msg);
-	else if (msg.is_method_call(INTERFACE, "CreateSingleSnapshot"))
-	    create_single_snapshot(conn, msg);
-	else if (msg.is_method_call(INTERFACE, "CreateSingleSnapshotV2"))
-	    create_single_snapshot_v2(conn, msg);
-	else if (msg.is_method_call(INTERFACE, "CreateSingleSnapshotOfDefault"))
-	    create_single_snapshot_of_default(conn, msg);
-	else if (msg.is_method_call(INTERFACE, "CreatePreSnapshot"))
-	    create_pre_snapshot(conn, msg);
-	else if (msg.is_method_call(INTERFACE, "CreatePostSnapshot"))
-	    create_post_snapshot(conn, msg);
-	else if (msg.is_method_call(INTERFACE, "DeleteSnapshots"))
-	    delete_snapshots(conn, msg);
-	else if (msg.is_method_call(INTERFACE, "GetDefaultSnapshot"))
-	    get_default_snapshot(conn, msg);
-	else if (msg.is_method_call(INTERFACE, "GetActiveSnapshot"))
-	    get_active_snapshot(conn, msg);
-	else if (msg.is_method_call(INTERFACE, "CalculateUsedSpace"))
-	    calculate_used_space(conn, msg);
-	else if (msg.is_method_call(INTERFACE, "GetUsedSpace"))
-	    get_used_space(conn, msg);
-	else if (msg.is_method_call(INTERFACE, "MountSnapshot"))
-	    mount_snapshot(conn, msg);
-	else if (msg.is_method_call(INTERFACE, "UmountSnapshot"))
-	    umount_snapshot(conn, msg);
-	else if (msg.is_method_call(INTERFACE, "GetMountPoint"))
-	    get_mount_point(conn, msg);
-	else if (msg.is_method_call(INTERFACE, "CreateComparison"))
-	    create_comparison(conn, msg);
-	else if (msg.is_method_call(INTERFACE, "DeleteComparison"))
-	    delete_comparison(conn, msg);
-	else if (msg.is_method_call(INTERFACE, "GetFiles"))
-	    get_files(conn, msg);
-	else if (msg.is_method_call(INTERFACE, "GetFilesByPipe"))
-	    get_files_by_pipe(conn, msg);
-	else if (msg.is_method_call(INTERFACE, "SetupQuota"))
-	    setup_quota(conn, msg);
-	else if (msg.is_method_call(INTERFACE, "PrepareQuota"))
-	    prepare_quota(conn, msg);
-	else if (msg.is_method_call(INTERFACE, "QueryQuota"))
-	    query_quota(conn, msg);
-	else if (msg.is_method_call(INTERFACE, "QueryFreeSpace"))
-	    query_free_space(conn, msg);
-	else if (msg.is_method_call(INTERFACE, "Sync"))
-	    sync(conn, msg);
-	else if (msg.is_method_call(INTERFACE, "Debug"))
-	    debug(conn, msg);
-	else
+	for (const vector<pair<const char*, method_fnc>>::value_type& tmp : method_registry)
 	{
-	    DBus::MessageError reply(msg, "error.unknown_method", DBUS_ERROR_FAILED);
-	    conn.send(reply);
+	    if (msg.is_method_call(INTERFACE, tmp.first))
+	    {
+		(*this.*tmp.second)(conn, msg);
+		return;
+	    }
 	}
+
+	DBus::MessageError reply(msg, "error.unknown_method", DBUS_ERROR_FAILED);
+	conn.send(reply);
+	return;
     }
     catch (const boost::thread_interrupted&)
     {
